/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
package net.sf.jmatchparser.util.charset.juniversalchardet;

import java.io.ByteArrayOutputStream;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;
import java.nio.charset.UnsupportedCharsetException;

import org.mozilla.universalchardet.UniversalDetector;

class JUniversalChardetCharset extends Charset {

	public static final String NAME = "jUniversalChardet";

	private static JUniversalChardetCharset instance = null;

	protected static JUniversalChardetCharset getInstance() {
		if (instance == null) {
			instance = new JUniversalChardetCharset();
		}
		return instance;
	}

	private JUniversalChardetCharset() {
		super(NAME, null);
	}

	@Override
	public boolean contains(Charset cs) {
		return cs == this;
	}

	@Override
	public CharsetDecoder newDecoder() {
		return new Decoder();
	}

	@Override
	public CharsetEncoder newEncoder() {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean canEncode() {
		return false;
	}

	private class Decoder extends CharsetDecoder {
		private UniversalDetector detector;
		private ByteArrayOutputStream buffer = new ByteArrayOutputStream();

		private CharsetDecoder usedDecoder = null;
		private ByteBuffer remaining = null;
		private boolean isFlushed = false;

		protected Decoder() {
			super(JUniversalChardetCharset.this, 1.0f, 2.0f);
			detector = new UniversalDetector(null);
		}

		private void setCharset(String charset) {
			try {
				usedDecoder = Charset.forName(charset).newDecoder();
			} catch (UnsupportedCharsetException ex) {
				usedDecoder = Charset.forName("ISO-8859-1").newDecoder();
			}
			usedDecoder.onUnmappableCharacter(unmappableCharacterAction());
			usedDecoder.onMalformedInput(malformedInputAction());
			byte[] buf = buffer.toByteArray();
			if (buf.length > 0) {
				remaining = ByteBuffer.allocate(buf.length);
				remaining.put(buf);
				remaining.flip();
			}
			buffer = null;
			detector = null;
		}

		@Override
		protected CoderResult decodeLoop(ByteBuffer in, CharBuffer out) {
			byte[] buf = new byte[1024];
			while (true) {
				if (usedDecoder != null) {
					if (remaining != null) {
						CoderResult result = writeOut(out);
						if (!result.isUnderflow())
							return result;
					}
					return usedDecoder.decode(in, out, false);
				}
				if (in.remaining() == 0)
					return CoderResult.UNDERFLOW;
				int len = Math.min(in.remaining(), buf.length);
				in.get(buf, 0, len);
				buffer.write(buf, 0, len);
				detector.handleData(buf, 0, len);
				if (detector.isDone()) {
					detector.dataEnd();
					setCharset(detector.getDetectedCharset());
				}
			}
		}

		@Override
		protected CoderResult implFlush(CharBuffer out) {
			if (usedDecoder == null) {
				detector.dataEnd();
				setCharset(detector.getDetectedCharset() == null ? "ISO-8859-1" : detector.getDetectedCharset());
			}
			if (remaining != null) {
				CoderResult result = writeOut(out);
				if (!result.isUnderflow())
					return result;
			}
			if (!isFlushed) {
				ByteBuffer empty = ByteBuffer.allocate(1);
				empty.flip();
				CoderResult result = usedDecoder.decode(empty, out, true);
				if (!result.isUnderflow())
					return result;
				result = usedDecoder.flush(out);
				if (!result.isUnderflow())
					return result;
				isFlushed = true;
			}
			return super.implFlush(out);
		}

		private CoderResult writeOut(CharBuffer out) {
			CoderResult result = usedDecoder.decode(remaining, out, false);
			if (remaining.remaining() == 0)
				remaining = null;
			return result;
		}

		@Override
		protected void implReset() {
			detector = new UniversalDetector(null);
			buffer = new ByteArrayOutputStream();
			usedDecoder = null;
			remaining = null;
			isFlushed = false;
		}

		@Override
		public boolean isAutoDetecting() {
			return true;
		}

		@Override
		public boolean isCharsetDetected() {
			return usedDecoder != null;
		}

		@Override
		public Charset detectedCharset() {
			return usedDecoder.charset();
		}
	}
}